/* Move down content because we have a fixed navbar that is 50px tall with 20px padding */
body {
  padding-top: 70px;
  padding-bottom: 20px;
}

.state_indicator {
  padding: 0 4px 0 4px;
}

.literal_output td {
  font-family: monospace;
}

.cursor-pointer {
  cursor: pointer;
}

.tooltip-inner {
  max-width: none;
  text-align: left;
}

.label {
  white-space: normal;
}

/* The navbar adds horizontal padding already */
.navbar .container-fluid {
    padding: 0;
}

/* This class provides style for containers that hold card like (without background) objects within container-fluid and out of accordion */
.blank-card {
  padding-bottom: 50px;
  padding-right: 25px;
}
